from queue import Queue

import pytest


@pytest.mark.parametrize(
    "fmt, representation",
    [
        ("po", "x2 = com.apple.powerlog.state_changed"),
        ("cf", "x2 = <CFString"),
        ("x", "x2 = 0x"),
        (lambda client, value: "Just return this string", "x2 = Just return this string"),
    ],
)
def test_monitor_register_formats(hilda_client, fmt: str, representation: str):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param fmt: Register format.
    :param representation: Substring of register's representation.
    """
    if "executable = aggregated" not in str(hilda_client.process):
        pytest.skip("Unsupported process, please run this test on aggregated")

    logs_queue = Queue()
    bp = hilda_client.symbols._PLStateChanged.monitor(regs={"x2": fmt}, stop=True)
    hilda_client.log_info = logs_queue.put_nowait
    # Trigger a call to _PLStateChanged.
    hilda_client.po("""
        @import ObjectiveC;
        @import Foundation;
        CFNotificationCenterRef center = CFNotificationCenterGetDarwinNotifyCenter();
        CFNotificationCenterPostNotification(center, CFSTR("com.apple.powerlog.state_changed"), NULL, NULL, FALSE);
    """)
    hilda_client.cont()
    output = logs_queue.get(timeout=10)
    hilda_client.breakpoints.remove(bp)
    assert representation in output



import logging

import inquirer3
from construct import Array, CString, Hex, If, Int32ub, Int32ul, Int64ul, Pointer, Struct, Tell, this
from inquirer3 import List

from hilda.lldb_importer import lldb
from hilda.snippets.macho.image_info import ImageInfo, dyld_image_info_t
from hilda.snippets.macho.macho import mach_header_t
from hilda.snippets.uuid import uuid_t
from hilda.symbol import SymbolFormatField

all_image_infos_t = Struct(
    "address" / Tell,
    "version" / Int32ul,
    "infoArrayCount" / Int32ul,
    "_infoArray" / SymbolFormatField(lldb.hilda_client),
    "infoArray" / Pointer(this._infoArray, Array(this.infoArrayCount, dyld_image_info_t)),
    "notification" / SymbolFormatField(lldb.hilda_client),
    "processDetachedFromSharedRegion" / Int32ub,
    "libSystemInitialized" / Int32ub,
    "_dyldImageLoadAddress" / SymbolFormatField(lldb.hilda_client),
    "dyldImageLoadAddress" / Pointer(this._dyldImageLoadAddress, mach_header_t),
    "jitInfo" / SymbolFormatField(lldb.hilda_client),
    "_dyldVersion" / SymbolFormatField(lldb.hilda_client),
    "dyldVersion" / Pointer(this._dyldVersion, CString("utf8")),
    "_errorMessage" / SymbolFormatField(lldb.hilda_client),
    "errorMessage" / If(this._errorMessage != 0, Pointer(this._erroressage, CString("utf8"))),
    "terminationFlags" / Int64ul,
    "coreSymbolicationShmPage" / SymbolFormatField(lldb.hilda_client),
    "systemOrderFlags" / Int64ul,
    "uuidArrayCount" / Int64ul,
    "_uuidArray" / SymbolFormatField(lldb.hilda_client),
    "uuidArray" / If(this._uuidArray != 0, Pointer(this._uuidArray, Array(this.uuidArrayCount, uuid_t))),
    "_dyld_all_image_infos" / SymbolFormatField(lldb.hilda_client),
    # 'dyld_all_image_infos' / Lazy(LazyBound(lambda: Pointer(this._dyld_all_image_infos, self.all_image_infos))),
    "initialImageCount" / Int64ul,
    "errorKind" / Int64ul,
    "_errorClientOfDyLibPath" / SymbolFormatField(lldb.hilda_client),
    "errorClientOfDyLibPath"
    / If(this._errorClientOfDyLibPath != 0, Pointer(this._errorClientofoyLibPath, CString("utf8"))),
    "_errorTargetDylibPath" / SymbolFormatField(lldb.hilda_client),
    "errorTargetDylibPath" / If(this._errorTargetDylibPath != 0, Pointer(this._errorTargetDylibPath, CString("utf8"))),
    "_errorSymbol" / SymbolFormatField(lldb.hilda_client),
    "errorSymbol" / If(this._errorSymbol != 0, Pointer(this._errorSymbol, CString("utf8"))),
    "sharedCacheSlide" / Hex(Int64ul),
)


class AllImageInfos:
    def reload(self):
        with lldb.hilda_client.stopped(1):
            all_image_infos_symbol = lldb.hilda_client.symbol(lldb.hilda_client.symbols.dyld_all_image_infos)
            self.__all_image_infos = all_image_infos_t.parse_stream(all_image_infos_symbol)
            self.__image_infos = []

            for image_info_data in self.__all_image_infos.infoArray:
                image_info = ImageInfo(image_info_data)
                self.__image_infos.append(image_info)

    def __init__(self):
        self.__all_image_infos = None
        self.__image_infos = None

        self.reload()

    @property
    def images(self):
        return self.__image_infos

    def find_images(self, name):
        return [image for image in self.images if name in image.file_path]

    def image_dependencies(self, image_name):
        images = list(self.find_images(image_name))

        if len(images) == 0:
            # This is a time consuming action. We only do it in
            # because a library could be dynamically opened
            self.reload()
            images = list(self.find_images(image_name))

            if len(images) == 0:
                print(f"Image {image_name} is not loaded")
                return

        return AllImageInfos.__image_dependencies(images)

    @staticmethod
    def __image_dependencies(images):
        unique_images = []

        for image in images:
            if image.file_path not in [unique_image.file_path for unique_image in unique_images]:
                unique_images.append(image)

        image = unique_images[0]
        if len(unique_images) > 1:
            image = AllImageInfos.__select_specific_image(unique_images)

        dependencies = {image_name.path for image_name in image.load_commands.dylib_commands}

        return dependencies

    @staticmethod
    def __select_specific_image(images):
        image_list = [image.file_path for image in images]
        logging.info(f"Multiple images were found with that prefix.\nPlease select one:\n{image_list}")
        selected_image_path = inquirer3.prompt(List("selected_image_path", image_list))["selected_image_path"]
        for image in images:
            if image.file_path == selected_image_path:
                return image



import pytest

from hilda.objective_c_class import PropertyAttributes, convert_encoded_property_attributes


def test_calling_class_method_by_method_name(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    NSDictionary = hilda_client.objc_get_class("NSDictionary")
    dictionary = NSDictionary.dictionaryWithObjects_forKeys_count_(0, 0, 0)
    assert dictionary.po() == "{\n}"
    dictionary = NSDictionary["dictionaryWithObjects:forKeys:count:"](0, 0, 0)
    assert dictionary.po() == "{\n}"


def test_calling_class_method_via_objc_call(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    NSDictionary = hilda_client.objc_get_class("NSDictionary")
    dictionary = NSDictionary.objc_call("dictionaryWithObjects:forKeys:count:", 0, 0, 0)
    assert dictionary.po() == "{\n}"


def test_getting_class_method(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    NSDictionary = hilda_client.objc_get_class("NSDictionary")
    assert "dictionaryWithObjects_forKeys_count_" in dir(NSDictionary)


def test_calling_super_class_method(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    NSDictionary = hilda_client.objc_get_class("NSDictionary")
    dictionary = NSDictionary.new()
    assert dictionary.po() == "{\n}"


def test_getting_super_class_method(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    NSDictionary = hilda_client.objc_get_class("NSDictionary")
    assert "new" in dir(NSDictionary)


@pytest.mark.parametrize(
    "encoded, result",
    [
        ("Tc,VcharDefault", PropertyAttributes(synthesize="charDefault", type_="char", list=[])),
        (
            "T{YorkshireTeaStruct=ic},VstructDefault",
            PropertyAttributes(
                synthesize="structDefault", type_="struct YorkshireTeaStruct { int x0; char x1; }", list=[]
            ),
        ),
        (
            "T@,R,&,VidReadonlyRetainNonatomic",
            PropertyAttributes(synthesize="idReadonlyRetainNonatomic", type_="id", list=["readonly", "strong"]),
        ),
        ("T^v", PropertyAttributes(synthesize=None, type_="void *", list=[])),
    ],
)
def test_convert_encoded_property_attributes(encoded: str, result: PropertyAttributes):
    """
    :param encoded: Property encoding.
    :param result: Parsed property data.
    """
    assert convert_encoded_property_attributes(encoded) == result


@pytest.mark.parametrize(
    "sub_str",
    [
        "@interface NSObject",
        "+ alloc;",
        "@property (readonly,copy) NSString * description;",
    ],
)
def test_class_without_super_str(hilda_client, sub_str: str):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param sub_str: Substring that should appear in the str.
    """
    assert str(hilda_client.objc_get_class("NSObject")).count(sub_str) == 1



from dataclasses import dataclass
from typing import Callable

from prompt_toolkit.enums import DEFAULT_BUFFER
from prompt_toolkit.filters import EmacsInsertMode, HasFocus, HasSelection, ViInsertMode
from prompt_toolkit.keys import Keys


@dataclass
class Keybinding:
    key: str
    description: str
    callback: Callable


def get_keybindings(hilda_client) -> list[Keybinding]:
    """
    Get list of keybindings

    :param hilda.hilda_client.HildaClient hilda_client: Hilda client to bind the keys to operations
    """
    return [
        Keybinding(key=Keys.F1, description="Show this help", callback=hilda_client.show_help),
        Keybinding(key=Keys.F2, description="Show process state UI", callback=hilda_client.ui_manager.show),
        Keybinding(
            key=Keys.F3,
            description="Toggle enabling of stdout & stderr",
            callback=hilda_client.toggle_enable_stdout_stderr,
        ),
        Keybinding(key=Keys.F7, description="Step Into", callback=hilda_client.step_into),
        Keybinding(key=Keys.F8, description="Step Over", callback=hilda_client.step_over),
        Keybinding(
            key=Keys.F9,
            description="Continue",
            callback=lambda _: (hilda_client.log_info("Sending continue"), hilda_client.cont()),
        ),
        Keybinding(
            key=Keys.F10,
            description="Stop",
            callback=lambda _: (hilda_client.log_info("Sending stop"), hilda_client.stop()),
        ),
    ]


def load_ipython_extension(ipython):
    def register_keybindings():
        hilda_client = ipython.user_ns["p"]
        insert_mode = ViInsertMode() | EmacsInsertMode()
        registry = ipython.pt_app.key_bindings

        for keybind in get_keybindings(hilda_client):
            registry.add_binding(keybind.key, filter=(HasFocus(DEFAULT_BUFFER) & ~HasSelection() & insert_mode))(
                keybind.callback
            )

    register_keybindings()
    ipython.events.register("shell_initialized", register_keybindings)



from construct import Computed, Hex, Int32ul, LazyArray, Struct, Tell, this

from hilda.snippets.macho.macho_load_commands import LOAD_COMMAND_TYPE, load_command_t


def __calculate_aslr(ctx):
    loaded_address = ctx.load_address
    for load_command in ctx.load_commands:
        if load_command.cmd == LOAD_COMMAND_TYPE.LC_SEGMENT_64 and load_command.data.segname == "__TEXT":
            return loaded_address - load_command.data.vmaddr


mach_header_t = Struct(
    "load_address" / Tell,
    "magic" / Hex(Int32ul),
    "cputype" / Hex(Int32ul),
    "cpusubtype" / Hex(Int32ul),
    "filetype" / Hex(Int32ul),
    "ncmds" / Hex(Int32ul),
    "sizeofcmds" / Hex(Int32ul),
    "flags" / Hex(Int32ul),
    "reserved" / Hex(Int32ul),
    "load_commands" / LazyArray(this.ncmds, load_command_t),
    "aslr" / Computed(__calculate_aslr),
)



from datetime import datetime, timezone

import pytest

from hilda.exceptions import ConvertingToNsObjectError


@pytest.mark.parametrize(
    "source",
    [
        "",
        "3123123",
        "asdsdasd",
        "12312sadasd",
        "The quick brown fox jumps over the lazy frog123",
    ],
)
def test_cfstr(hilda_client, source: str):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param source: Python string to be converted to CFSTR.
    """
    cfstr = hilda_client.CFSTR(source)
    assert cfstr
    po = cfstr.po()
    assert po == source or (not source and po == "<object returned empty description>")
    assert source in cfstr.cf_description
    assert "NSString" in [sup.name for sup in cfstr.objc_class.iter_supers()]


def test_ns_data(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    data = b"\x01\x00asdasd\xff\xfe58"
    ns_data = hilda_client.ns(data).objc_symbol
    assert ns_data.bytes.read(len(data)) == data


@pytest.mark.parametrize("day, month, year", [(1, 1, 1970), (11, 10, 2021)])
def test_ns_date(hilda_client, day: int, month: int, year: int):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param day: Date day.
    :param month: Date month.
    :param year: Date year.
    """
    date_po = f"{year:04d}-{month:02d}-{day:02d} 00:00:00 +0000"
    assert hilda_client.ns(datetime(day=day, month=month, year=year, tzinfo=timezone.utc)).po() == date_po


@pytest.mark.parametrize(
    "source, result",
    [
        (
            {"asdasd": 234234, 234234: "asdasd", 1: True, "a": [False, False]},
            "{1=1;234234=asdasd;a=(0,0);asdasd=234234;}",
        ),
        ({}, "{}"),
        ({"asdasds": 324234}, "{asdasds=324234;}"),
        ({1: {2: {3: "a"}}}, "{1={2={3=a;};};}"),
        ({1: b"\x00\x01"}, "{1={length=2,bytes=0x0001};}"),
        ({1: datetime(1970, 1, 1, tzinfo=timezone.utc)}, '{1="1970-01-0100:00:00+0000";}'),
    ],
)
def test_ns_dict(hilda_client, source: dict, result: str):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param source: Python object to be converted to NS object.
    :param result: PO result.
    """
    ns_object = hilda_client.ns(source)
    assert "".join(ns_object.po().split()) == result
    assert "NSDictionary" in [sup.name for sup in ns_object.objc_class.iter_supers()]


def test_ns_none(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    ns_object = hilda_client.ns(None)
    assert ns_object.po() == "<null>"
    assert ns_object.objc_class.name == "NSNull"


@pytest.mark.parametrize(
    "source, result",
    [
        (["asdasd", 234234, 1, True, {"a": False}], "(asdasd,234234,1,1,{a=0;})"),
        ([], "()"),
        (["asdasds", 324234], "(asdasds,324234)"),
        ([1, [2, [3, "a"]]], "(1,(2,(3,a)))"),
        (["asdasds", b"324234"], "(asdasds,{length=6,bytes=0x333234323334})"),
        (["asdasds", datetime(2014, 4, 1, tzinfo=timezone.utc)], '(asdasds,"2014-04-0100:00:00+0000")'),
    ],
)
def test_ns_array(hilda_client, source, result: str):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param source: Python object to be converted to NS object.
    :param result: PO result.
    """
    ns_object = hilda_client.ns(source)
    assert "".join(ns_object.cf_description.split()) == result
    assert "NSArray" in [sup.name for sup in ns_object.objc_class.iter_supers()]


@pytest.mark.parametrize("source", [0, 1, -1, 1.5, -1.5, 0xFFFFFFFFFFFFFFFFFFF, -0xFFFFFFFFFFFFFFFFFFF, 1 / 3])
def test_ns_number(hilda_client, source):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param source: Number to convert to NS object.
    """
    ns_object = hilda_client.ns(source)
    assert ns_object.po() == str(source)
    assert "NSNumber" in [sup.name for sup in ns_object.objc_class.iter_supers()]


@pytest.mark.parametrize(
    "source",
    [
        object(),
        {"aa", 123},
    ],
)
def test_error_converting_to_ns(hilda_client, source):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param source: Python object to convert to NS object.
    """
    with pytest.raises(ConvertingToNsObjectError):
        hilda_client.ns(source)



import json
import time
from collections import UserList, namedtuple
from dataclasses import dataclass, field
from functools import partial
from pathlib import Path
from typing import Any, Optional

from objc_types_decoder.decode import decode as decode_type
from objc_types_decoder.decode import decode_with_tail
from pygments import highlight
from pygments.formatters import TerminalTrueColorFormatter
from pygments.lexers import ObjectiveCLexer

from hilda.exceptions import GettingObjectiveCClassError

Ivar = namedtuple("Ivar", "name type_ offset")
Property = namedtuple("Property", "name attributes")
PropertyAttributes = namedtuple("PropertyAttributes", "synthesize type_ list")


def convert_encoded_property_attributes(encoded):
    conversions = {
        "R": lambda x: "readonly",
        "C": lambda x: "copy",
        "&": lambda x: "strong",
        "N": lambda x: "nonatomic",
        "G": lambda x: "getter=" + x[1:],
        "S": lambda x: "setter=" + x[1:],
        "d": lambda x: "dynamic",
        "W": lambda x: "weak",
        "P": lambda x: "<garbage-collected>",
        "t": lambda x: "encoding=" + x[1:],
    }

    type_, tail = decode_with_tail(encoded[1:])
    attributes = []
    synthesize = None
    for attr in filter(None, tail.lstrip(",").split(",")):
        if attr[0] in conversions:
            attributes.append(conversions[attr[0]](attr))
        elif attr[0] == "V":
            synthesize = attr[1:]

    return PropertyAttributes(type_=type_, synthesize=synthesize, list=attributes)


@dataclass
class Method:
    name: str
    client: Any = field(compare=False)
    address: int = field(compare=False)
    imp: int = field(compare=False)
    type_: str = field(compare=False)
    return_type: str = field(compare=False)
    is_class: bool = field(compare=False)
    args_types: list = field(compare=False)
    class_name: str = field(compare=False)

    @staticmethod
    def from_data(class_name: str, data: dict, client) -> "Method":
        """
        Create Method object from raw data.
        :param class_name: ObjC class name
        :param data: Data as loaded from get_objectivec_symbol_data.m.
        :param hilda.hilda_client.HildaClient client: Hilda client.
        """
        return Method(
            name=data["name"],
            client=client,
            address=client.symbol(data["address"]),
            imp=client.symbol(data["imp"]),
            type_=data["type"],
            return_type=decode_type(data["return_type"]),
            is_class=data["is_class"],
            args_types=list(map(decode_type, data["args_types"])),
            class_name=class_name,
        )

    def set_implementation(self, new_imp: int):
        self.client.symbols.method_setImplementation(self.address, new_imp)
        self.imp = self.client.symbol(new_imp)

    def monitor(self, **args) -> None:
        """
        Perform monitor on method's IMP.
        See monitor command for more details.
        :param args: given arguments for monitor command
        """
        self.client.symbol(self.imp).monitor(**args)

    def bp(self, **args) -> None:
        """
        Place a breakpoint on method's IMP.
        See bp command for more details.
        :param args: given arguments for bp command
        """
        self.client.symbol(self.imp).bp(**args)

    def __str__(self) -> str:
        if ":" in self.name:
            args_names = self.name.split(":")
            name = " ".join(["{}:({})".format(*arg) for arg in zip(args_names, self.args_types[2:])])
        else:
            name = self.name
        prefix = "+" if self.is_class else "-"
        return f"{prefix} {name}; // 0x{self.address:x} (returns: {self.return_type})\n"


class MethodList(UserList):
    def __init__(self, class_name: str, methods: list[Method]) -> None:
        super().__init__()
        self._class_name = class_name
        self.data = methods

    def get(self, name: str) -> Optional[Method]:
        """
        Get a specific method implementation.
        :param name: Method name.
        :return: Method.
        """
        for method in self.data:
            if method.name == name:
                return method
        return None

    # Actions

    def bp(self, callback=None, **kwargs):
        """
        Place a breakpoint on all symbols in the method list.
        Look for the bp command for more details.
        :param callback:  callback function to be executed upon an hit
        :param kwargs: optional kwargs for the bp command
        """
        for method in self.data:
            kwargs["name"] = f"[{self._class_name} {method.name}]"
            method.imp.bp(callback, **kwargs)

    def monitor(self, **kwargs):
        """
        Perform monitor for all symbols in the method list.
        See monitor command for more details.
        :param args: given arguments for monitor command
        """
        for method in self.data:
            method_kwargs = kwargs.copy()
            method_kwargs["name"] = f"{'+' if method.is_class else '-'}[{method.class_name} {method.name}]"
            method.imp.monitor(**method_kwargs)

    # Filters

    def filter_startswith(self, exp, case_sensitive=True):
        """
        Filter only methods with given prefix
        :param exp: prefix
        :param case_sensitive: is case sensitive
        :return: reduced method list
        """
        if not case_sensitive:
            exp = exp.lower()

        retval = []
        for v in self.data:
            name = v.name
            if not case_sensitive:
                name = name.lower()
            if name.startswith(exp):
                retval.append(v)
        return MethodList(self._class_name, retval)

    def filter_name_contains(self, exp, case_sensitive=True):
        """
        Filter methods containing a given expression
        :param exp: given expression
        :param case_sensitive: is case sensitive
        :return: reduced method list
        """
        if not case_sensitive:
            exp = exp.lower()

        retval = []
        for v in self.data:
            name = v.name
            if not case_sensitive:
                name = name.lower()
            if exp in name:
                retval.append(v)
        return MethodList(self._class_name, retval)


class Class:
    """
    Wrapper for ObjectiveC Class object.
    """

    def __init__(self, client, class_object=0, class_data: Optional[dict] = None):
        """
        :param hilda.hilda_client.HildaClient client:
        :param hilda.symbol.Symbol class_object:
        """
        self._client = client
        self._class_object = class_object
        self.protocols = []
        self.ivars = []
        self.properties = []
        self.name = ""
        self.methods = MethodList(self.name, [])
        self.super = None
        if class_data is None:
            self.reload()
        else:
            self._load_class_data(class_data)

    @staticmethod
    def from_class_name(client, class_name: str):
        """
        Create ObjectiveC Class from given class name.
        :param hilda.hilda_client.HildaClient client: Hilda client.
        :param class_name: Class name.
        """
        obj_c_code = (client._hilda_root / "objective_c" / "get_objectivec_class_description.m").read_text()
        obj_c_code = obj_c_code.replace("__class_address__", "0").replace("__class_name__", class_name)
        class_symbol = Class(client, class_data=json.loads(client.po(obj_c_code)))
        if class_symbol.name != class_name:
            raise GettingObjectiveCClassError()
        return class_symbol

    @staticmethod
    def sanitize_name(name: str):
        """
        Sanitize python name to ObjectiveC name.
        """
        name = "_" + name[1:].replace("_", ":") if name.startswith("_") else name.replace("_", ":")
        return name

    def reload(self):
        """
        Reload class object data.
        Should be used whenever the class layout changes (for example, during method swizzling)
        """
        obj_c_code = (self._client._hilda_root / "objective_c" / "get_objectivec_class_description.m").read_text()
        obj_c_code = obj_c_code.replace("__class_address__", f"{self._class_object:d}")
        obj_c_code = obj_c_code.replace("__class_name__", self.name)
        self._load_class_data(json.loads(self._client.po(obj_c_code)))

    def show(self):
        """
        Print to terminal the highlighted class description.
        """
        print(highlight(str(self), ObjectiveCLexer(), TerminalTrueColorFormatter(style="native")))

    def objc_call(self, sel: str, *args):
        """
        Invoke a selector on the given class object.
        :param sel: Selector name.
        :return: whatever the selector returned as a symbol.
        """
        return self._class_object.objc_call(sel, *args)

    def capture_self(self, sync: bool = False):
        """
        Capture the first called `self` from this class.
        Access using `self.captured_objects`
        :param sync: Should wait until captured object is returned?
        :return: Captured object if sync is True, None otherwise
        """
        class_name = self.name

        if class_name in self._client.captured_objects:
            del self._client.captured_objects[class_name]

        def hook(hilda, frame, bp_loc, options):
            hilda.log_info(f"self object has been captured from {options['name']}")
            hilda.log_info("removing breakpoints")
            for bp_id, bp in list(hilda.breakpoints.items()):
                if "group_uuid" in bp.options and bp.options.get("group_uuid", "") == options["group_uuid"]:
                    hilda.breakpoints.remove(bp_id)
            captured = hilda.evaluate_expression("$arg1")
            captured = captured.objc_symbol
            hilda.captured_objects[options["name"].split(" ")[0].split("[")[1]] = captured
            hilda.cont()

        for method in self.methods:
            if not method.is_class:
                # only instance methods are relevant for capturing self
                method.imp.bp(hook)

        if sync:
            self._client.cont()
            self._client.log_debug("Waiting for desired object to be captured...")
            while class_name not in self._client.captured_objects:
                time.sleep(1)

            return self._client.captured_objects[class_name]

    def monitor(self, **kwargs):
        """
        Proxy for monitor command.
        """
        self.methods.monitor(**kwargs)

    def bp(self, callback=None, **kwargs):
        """
        Proxy for bp command.
        """
        self.methods.bp(callback, **kwargs)

    def iter_supers(self):
        """
        Iterate over the super classes of the class.
        """
        sup = self.super
        while sup is not None:
            yield sup
            sup = sup.super

    @property
    def bundle_path(self) -> Path:
        return Path(
            self._client.symbols
            .objc_getClass("NSBundle")
            .objc_call("bundleForClass:", self._class_object)
            .objc_call("bundlePath")
            .py()
        )

    def _load_class_data(self, data: dict):
        self._class_object = self._client.symbol(data["address"])
        self.super = Class(self._client, data["super"]) if data["super"] else None
        self.name = data["name"]
        self.protocols = data["protocols"]
        self.ivars = [
            Ivar(
                name=ivar["name"],
                type_=decode_type(ivar["type"]) if ivar["type"] else "unknown_type_t",
                offset=ivar["offset"],
            )
            for ivar in data["ivars"]
        ]
        self.properties = [
            Property(name=prop["name"], attributes=convert_encoded_property_attributes(prop["attributes"]))
            for prop in data["properties"]
        ]
        self.methods = MethodList(
            self.name, [Method.from_data(self.name, method, self._client) for method in data["methods"]]
        )

    def __dir__(self):
        result = set()

        for method in self.methods:
            if method.is_class:
                result.add(method.name.replace(":", "_"))

        for sup in self.iter_supers():
            if self._client.configs.nsobject_exclusion and sup.name == "NSObject":
                continue
            for method in sup.methods:
                if method.is_class:
                    result.add(method.name.replace(":", "_"))

        result.update(list(super().__dir__()))
        return list(result)

    def __str__(self):
        protocol_buf = f"<{','.join(self.protocols)}>" if self.protocols else ""

        if self.super is not None:
            buf = f"@interface {self.name}: {self.super.name} {protocol_buf}\n"
        else:
            buf = f"@interface {self.name} {protocol_buf}\n"

        # Add ivars
        buf += "{\n"
        for ivar in self.ivars:
            buf += f"\t{ivar.type_} {ivar.name}; // 0x{ivar.offset:x}\n"
        buf += "}\n"

        # Add properties
        for prop in self.properties:
            buf += f"@property ({','.join(prop.attributes.list)}) {prop.attributes.type_} {prop.name};\n"

            if prop.attributes.synthesize is not None:
                buf += f"@synthesize {prop.name} = {prop.attributes.synthesize};\n"

        # Add methods
        for method in self.methods:
            buf += str(method)

        buf += "@end"
        return buf

    def __repr__(self):
        return f'<objC Class "{self.name}">'

    def __getitem__(self, item):
        for method in self.methods:
            if method.name == item:
                if method.is_class:
                    return partial(self.objc_call, item)
                else:
                    raise AttributeError(f"{self.name} class has an instance method named {item}, not a class method")

        for sup in self.iter_supers():
            for method in sup.methods:
                if method.name == item:
                    if method.is_class:
                        return partial(self.objc_call, item)
                    else:
                        raise AttributeError(
                            f"{self.name} class has an instance method named {item}, not a class method"
                        )

        raise AttributeError(f"""'{self.name}' class has no attribute {item}""")

    def __getattr__(self, item: str):
        return self[self.sanitize_name(item)]



from datetime import datetime, timezone

import pytest

from hilda.exceptions import ConvertingFromNSObjectError


@pytest.mark.parametrize(
    "source",
    [
        "",
        "3123123",
        "asdsdasd",
        "12312sadasd",
        "The quick brown fox jumps over the lazy frog123",
    ],
)
def test_cfstr(hilda_client, source: str):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param source: CFSTR to be converted to Python string.
    """
    cfstr = hilda_client.evaluate_expression(f'@"{source}"')
    assert cfstr
    assert hilda_client.decode_cf(cfstr) == source


def test_ns_data(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    data = b"\x01\x00asdasd\xff\xfe58"
    with hilda_client.safe_malloc(len(data)) as buffer:
        buffer.write(data)
        ns_data = hilda_client.evaluate_expression(f"[NSData dataWithBytes:(char *)0x{buffer:x} length:{len(data)}]")
        assert hilda_client.decode_cf(ns_data) == data


def test_ns_data_in_dict(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    data = b"\x01\x00asdasd\xff\xfe58"
    with hilda_client.safe_malloc(len(data)) as buffer:
        buffer.write(data)
        ns_dict = hilda_client.evaluate_expression(
            f'@{{"a": [NSData dataWithBytes:(char *)0x{buffer:x} length:{len(data)}]}}'
        )
        assert hilda_client.decode_cf(ns_dict) == {"a": data}


def test_ns_data_in_array(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    data = b"\x01\x00asdasd\xff\xfe58"
    with hilda_client.safe_malloc(len(data)) as buffer:
        buffer.write(data)
        ns_dict = hilda_client.evaluate_expression(f"@[[NSData dataWithBytes:(char *)0x{buffer:x} length:{len(data)}]]")
        assert hilda_client.decode_cf(ns_dict) == [data]


@pytest.mark.parametrize("day, month, year", [(1, 1, 1970), (11, 10, 2021)])
def test_ns_date(hilda_client, day: int, month: int, year: int):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param day: Date day.
    :param month: Date month.
    :param year: Date year.
    """
    date = hilda_client.evaluate_expression(f"""
        NSDateComponents *comps = [NSDateComponents new];
        [comps setDay:{day}];
        [comps setMonth:{month}];
        [comps setYear:{year}];
        [comps setTimeZone:[NSTimeZone timeZoneWithAbbreviation:@"UTC"]];
        [[NSCalendar currentCalendar] dateFromComponents:comps];
    """)
    assert hilda_client.decode_cf(date) == datetime(day=day, month=month, year=year, tzinfo=timezone.utc)


@pytest.mark.parametrize(
    "source, result",
    [
        # Dictionaries
        (
            '@{@1:@1,@234234:@"asdasd",@"a":@[@0,@0],@"asdasd":@234234}',
            {"asdasd": 234234, 234234: "asdasd", 1: 1, "a": [0, 0]},
        ),
        ("@{}", {}),
        ('@{"asdasds":324234}', {"asdasds": 324234}),
        ("@{[NSNull null]:324234}", {None: 324234}),
        ('@{@{"a":1}:324234}', {(("a", 1),): 324234}),
        ('@{@["a",1]:324234}', {("a", 1): 324234}),
        ('@{1:@{2:@{3:"a"}}}', {1: {2: {3: "a"}}}),
        # Arrays
        ('@[@"asdasd",@234234,@1,@1,@{@"a":@0}]', ["asdasd", 234234, 1, 1, {"a": 0}]),
        ("@[]", []),
        ('@["asdasds",324234]', ["asdasds", 324234]),
        ('@[1,@[2,@[3,"a"]]]', [1, [2, [3, "a"]]]),
        # Sets
        ('[NSSet setWithArray: @[@"1", @42]]', ["1", 42]),
    ],
)
def test_ns_nested_objects(hilda_client, source: str, result):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param source: NS object expression to be converted to Python object.
    :param result: Python object.
    """
    ns_object = hilda_client.evaluate_expression(source)
    assert ns_object
    assert hilda_client.decode_cf(ns_object) == result


def test_ns_none(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    ns_object = hilda_client.evaluate_expression("[NSNull null]")
    assert hilda_client.decode_cf(ns_object) is None


@pytest.mark.parametrize("source", [0, 1, -1, 1.5, -1.5, 0xFFFFFFFFFFFFFFFFFFF, -0xFFFFFFFFFFFFFFFFFFF, 1 / 3])
def test_ns_number(hilda_client, source):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param source: Number to convert from NS object.
    """
    ns_number = hilda_client.evaluate_expression(f'[NSDecimalNumber decimalNumberWithString:@"{source}"]')
    assert hilda_client.decode_cf(ns_number) == source


@pytest.mark.parametrize(
    "source",
    [
        "[NSObject new]",
        "0x33",
    ],
)
def test_error_converting_from_ns(hilda_client, source):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param source: Expression to convert to python object.
    """
    with pytest.raises(ConvertingFromNSObjectError):
        hilda_client.decode_cf(hilda_client.evaluate_expression(source))



import logging
from pathlib import Path
from typing import Optional

import click
import coloredlogs

from hilda import launch_lldb
from hilda._version import version
from hilda.launch_lldb import (
    create_hilda_client_using_attach_by_name,
    create_hilda_client_using_attach_by_pid,
    create_hilda_client_using_launch,
    create_hilda_client_using_remote_attach,
)

DEFAULT_HILDA_PORT = 1234

coloredlogs.install(level=logging.DEBUG)


@click.group()
def cli():
    pass


startup_files_option = click.option("-f", "--startup_files", multiple=True, help="Files to run on start")


def parse_envp(ctx: click.Context, param: click.Parameter, value: list[str]) -> list[str]:
    env_list = []
    for item in value:
        try:
            key, val = item.split("=", 1)
            env_list.append(f"{key}={val}")
        except ValueError as e:
            raise click.BadParameter(f"Invalid format for --envp: {item}. Expected KEY=VALUE.") from e
    return env_list


@cli.command("remote")
@click.argument("hostname", default="localhost")
@click.argument("port", type=click.INT, default=DEFAULT_HILDA_PORT)
@startup_files_option
def remote(hostname: str, port: int, startup_files: list[str]) -> None:
    """Connect to remote debugserver at given address"""
    with create_hilda_client_using_remote_attach(hostname, port) as hilda_client:
        hilda_client.interact(startup_files=startup_files)


@cli.command("attach")
@click.option("-n", "--name", help="process name to attach")
@click.option("-p", "--pid", type=click.INT, help="pid to attach")
@startup_files_option
def attach(name: Optional[str], pid: Optional[int], startup_files: list[str]) -> None:
    """Attach to given process and start a lldb shell"""
    if name is not None:
        hilda_client = create_hilda_client_using_attach_by_name(name)
    elif pid is not None:
        hilda_client = create_hilda_client_using_attach_by_pid(pid)
    else:
        raise click.UsageError("You must specify a process name or pid")
    hilda_client.interact(startup_files=startup_files)
    hilda_client.detach()


@cli.command("bare")
def cli_bare() -> None:
    """Just start a lldb shell"""
    commands = [f"command script import {Path(__file__).resolve().parent / 'lldb_entrypoint.py'}"]
    commands = "\n".join(commands)
    launch_lldb.execute(f'lldb --one-line "{commands}"')


@cli.command("launch")
@click.argument("exec_path")
@click.option("--argv", multiple=True, help="Command line arguments to pass to the process")
@click.option("--envp", multiple=True, callback=parse_envp, help="Environment variables in the form KEY=VALUE")
@click.option("--stdin", help="Redirect stdin from this file path")
@click.option("--stdout", help="Redirect stdout to this file path")
@click.option("--stderr", help="Redirect stderr to this file path")
@click.option("--cwd", help="Set the working directory for the process")
@click.option("--flags", type=click.INT, default=0, help="Launch flags (bitmask)")
@startup_files_option
def launch(
    exec_path: str,
    argv: list[str],
    envp: list[str],
    stdin: Optional[Path],
    stdout: Optional[Path],
    stderr: Optional[Path],
    cwd: Optional[Path],
    flags: Optional[int],
    startup_files: list[str],
) -> None:
    """Attach to a given process and start a lldb shell"""
    argv = list(argv)
    envp = list(envp)
    with create_hilda_client_using_launch(exec_path, argv, envp, stdin, stdout, stderr, cwd, flags) as hilda_client:
        hilda_client.interact(startup_files=startup_files)


@cli.command("version")
def cli_version() -> None:
    """Show the version information."""
    click.echo(version)



from hilda.lldb_importer import lldb

TLV_MAP = {
    0x00: "METHOD",
    0x01: "IDENTIFIER",
    0x02: "SALT",
    0x03: "PUBLIC_KEY",
    0x04: "PROOF",
    0x05: "ENCRYPTED_DATA",
    0x06: "STATE",
    0x07: "ERROR",
    0x08: "RETRY_DELAY",
    0x09: "CERTIFICATE",
    0x0A: "SIGNATURE",
    0x0B: "PERMISSIONS",
    0x0C: "FRAGMENT_DATA",
    0x0D: "FRAGMENT_LAST",
    0x0E: "SESSION_ID",
    0x0F: "TTL",
    0x10: "EXTRA_DATA",
    0x11: "INFO",
    0x12: "ACL",
    0x13: "FLAGS",
    0x14: "VALIDATION_DATA",
    0x15: "MFI_AUTH_TOKEN",
    0x16: "MFI_PRODUCT_TYPE",
    0x17: "SERIAL_NUMBER",
    0x18: "MFI_AUTH_TOKEN_UUID",
    0x19: "APP_FLAGS",
    0x1A: "OWNERSHIP_PROOF",
    0x1B: "SETUP_CODE_TYPE",
    0x1C: "PRODUCTION_DATA",
    0x1D: "APP_INFO",
    0xFF: "SEPARATOR",
}


def _TLV8CopyCoalesced_bp(hilda, *args):
    type_ = hilda.registers.x2
    out_len = hilda.registers.x3
    out_len.item_size = 4
    hilda.finish()

    if hilda.registers.x0 == 0:
        print(f"TLV8CopyCoalesced\n    type_: {TLV_MAP[type_]}\n    buffer: Null\n---\n")
    else:
        print(
            f"TLV8CopyCoalesced\n    type_: {TLV_MAP[type_]}\n    buffer: {hilda.registers.x0.peek(out_len[0])}\n---\n"
        )
    hilda.cont()


def _TLV8BufferAppend_bp(hilda, *args):
    buffer = hilda.registers.x0
    type_ = hilda.registers.x1
    buffer = hilda.registers.x2
    buffer_len = hilda.registers.x3
    print(f"TLV8BufferAppend\n    type_: {TLV_MAP[type_]}\n    buffer: {buffer.peek(buffer_len)}\n---\n")
    hilda.cont()


def _SRPClientStart_libsrp_bp(hilda, *args):
    username = hilda.registers.x2
    username_len = hilda.registers.x3
    password = hilda.registers.x4
    password_len = hilda.registers.x5
    salt = hilda.registers.x6
    salt_len = hilda.registers.x7
    server_public_key = hilda.registers.sp[0]
    server_public_key_len = hilda.registers.sp[1]
    client_public_key = hilda.registers.sp[2]
    client_public_key_len = hilda.registers.sp[3]
    print(
        f"SRPClientStart_libsrp\n"
        f"    username: {username.peek(username_len)}\n"
        f"    password: {password.peek(password_len)}\n"
        f"    salt: {salt.peek(salt_len)}\n"
        f"    server_public_key: {server_public_key.peek(server_public_key_len)}\n"
        f"    client_public_key: {client_public_key.peek(client_public_key_len)}\n"
        f"---\n"
    )
    hilda.cont()


def _cced25519_sign_bp(hilda, *args):
    sig = hilda.registers.x1
    len = hilda.registers.x2
    msg = hilda.registers.x3
    msg = msg.peek(len)
    public_key = hilda.registers.x4.peek(32)
    private_key = hilda.registers.x5.peek(32)
    hilda.finish()
    print(
        f"cced25519_sign\n"
        f"    public_key: {public_key}\n"
        f"    private_key: {private_key}\n"
        f"    msg: {msg}\n"
        f"    sig: {sig.peek(64)}\n"
        f"---\n"
    )
    hilda.cont()


def _cced25519_verify_bp(hilda, *args):
    sig = hilda.registers.x3
    len = hilda.registers.x1
    msg = hilda.registers.x2
    msg = msg.peek(len)
    public_key = hilda.registers.x4.peek(32)
    hilda.finish()
    print(f"cced25519_verify\n    public_key: {public_key}\n    msg: {msg}\n    sig: {sig.peek(64)}\n---\n")
    hilda.cont()


def _CryptoHKDF_bp(hilda, *args):
    descriptor = hilda.registers.x0
    key = hilda.registers.x1.peek(hilda.registers.x2)
    salt = hilda.registers.x3
    salt_len = hilda.registers.x4
    info = hilda.registers.x5
    info_len = hilda.registers.x6
    out_key = hilda.registers.sp[0]
    hilda.finish()
    print(
        f"CryptoHKDF\n"
        f"    descriptor: {descriptor}\n"
        f"    key: {key}\n"
        f"    salt: {salt.peek(salt_len)}\n"
        f"    info: {info.peek(info_len)}\n"
        f"    outKey: {out_key.peek(32)}\n"
        f"---\n"
    )
    hilda.cont()


def _chacha20_poly1305_encrypt_all_bp(hilda, *args):
    key = hilda.registers.x0.peek(32)
    nonce = hilda.registers.x1.peek(hilda.registers.x2)
    plaintext = hilda.registers.x5
    plaintext_len = hilda.registers.x6
    print(
        f"_chacha20_poly1305_encrypt_all\n"
        f"    key: {key}\n"
        f"    nonce: {nonce}\n"
        f"    plaintext: {plaintext.peek(plaintext_len)}\n"
        f"---\n"
    )
    hilda.cont()


def _chacha20_poly1305_decrypt_all_bp(hilda, *args):
    key = hilda.registers.x0.peek(32)
    nonce = hilda.registers.x1.peek(hilda.registers.x2)
    encrypted = hilda.registers.x5
    encrypted_len = hilda.registers.x6
    plaintext = hilda.registers.x7
    encrypted_buf = encrypted.peek(encrypted_len)
    hilda.finish()
    print(
        f"_chacha20_poly1305_decrypt_all\n"
        f"    key: {key}\n"
        f"    nonce: {nonce}\n"
        f"    encrypted: {encrypted_buf}\n"
        f"    plaintext: {plaintext.peek(encrypted_len)}\n"
        f"---\n"
    )
    hilda.cont()


def monitor_crypto_functions() -> None:
    hilda_client = lldb.hilda_client

    hilda_client.symbols.TLV8CopyCoalesced.bp(_TLV8CopyCoalesced_bp)
    hilda_client.symbols.TLV8BufferAppend.bp(_TLV8BufferAppend_bp)
    hilda_client.symbols.SRPClientStart_libsrp.bp(_SRPClientStart_libsrp_bp)
    hilda_client.symbols.SRPClientVerify_libsrp.monitor()
    hilda_client.symbols.SRPServerStart_libsrp.monitor()
    hilda_client.symbols.SRPServerVerify_libsrp.monitor()
    hilda_client.symbols.cced25519_sign.bp(_cced25519_sign_bp)
    hilda_client.symbols.cced25519_verify.bp(_cced25519_verify_bp)
    hilda_client.symbols.CryptoHKDF.bp(_CryptoHKDF_bp)
    hilda_client.symbols._chacha20_poly1305_encrypt_all.bp(_chacha20_poly1305_encrypt_all_bp)
    hilda_client.symbols._chacha20_poly1305_decrypt_all.bp(_chacha20_poly1305_decrypt_all_bp)



import pytest

from hilda.exceptions import CreatingObjectiveCSymbolError
from hilda.objective_c_symbol import ObjectiveCSymbol


@pytest.mark.parametrize(
    "literal, result",
    [
        ('@{@"one": @1}', "{\n    one = 1;\n}"),
        ("@{}", "{\n}"),
    ],
)
def test_create_objective_c_symbol(hilda_client, literal: str, result: str):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param literal: ObjectiveC object literal.
    :param result: Object po.
    """
    assert hilda_client.evaluate_expression(literal).objc_symbol.po() == result


def test_fail_create_objective_c_symbol(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    with pytest.raises(CreatingObjectiveCSymbolError):
        hilda_client.symbol(1).objc_symbol


def test_access_missing_attribute(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    single_key_dictionary = hilda_client.evaluate_expression('@{@"one": @1}').objc_symbol
    with pytest.raises(AttributeError):
        assert single_key_dictionary.attribute_that_doesnt_exist


def test_get_objective_c_ivar(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    single_key_dictionary = hilda_client.evaluate_expression('@{@"one": @1}').objc_symbol
    assert single_key_dictionary._key.po() == "one"


def test_change_objective_c_ivar(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    single_key_dictionary = hilda_client.evaluate_expression('@{@"one": @1}').objc_symbol
    single_key_dictionary._key = hilda_client.ns("two")
    assert single_key_dictionary.description.po() == "{\n    two = 1;\n}"
    assert single_key_dictionary._key.po() == "two"


def test_call_objective_c_property(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    single_key_dictionary = hilda_client.evaluate_expression('@{@"one": @1}').objc_symbol
    assert single_key_dictionary.description.po() == "{\n    one = 1;\n}"


def test_call_objective_c_method_by_method_name(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    dictionary1 = hilda_client.evaluate_expression('@{@"one": @1}').objc_symbol
    dictionary2 = hilda_client.evaluate_expression('@{@"one": @2}').objc_symbol
    assert dictionary1.isEqualToDictionary_(dictionary1)
    assert not dictionary1.isEqualToDictionary_(dictionary2)


def test_call_objective_c_method_by_objc_call(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    dictionary1 = hilda_client.evaluate_expression('@{@"one": @1}').objc_symbol
    dictionary2 = hilda_client.evaluate_expression('@{@"one": @2}').objc_symbol
    assert dictionary1.objc_call("isEqualToDictionary:", dictionary1)
    assert not dictionary1.objc_call("isEqualToDictionary:", dictionary2)


def test_call_objective_c_method_returns_native_symbol(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    string1 = hilda_client.ns("asdasd").objc_symbol
    assert not isinstance(string1.UTF8String(), ObjectiveCSymbol)


@pytest.mark.parametrize(
    "selector",
    [
        "alloc",  # Test super class methods.
        "init",  # Test super instance methods.
        "dictionaryWithObjects_forKeys_",  # Test class methods.
        "objectForKey_",  # Test instance methods.
        "allKeys",  # Test instance properties.
    ],
)
def test_symbol_dir(hilda_client, selector: str):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param selector: Superclass selector to test.
    """
    dict_dir = dir(hilda_client.ns({1: 2}).objc_symbol)
    assert dict_dir.count(selector) == 1


@pytest.mark.parametrize(
    "sub_str",
    [
        "@interface NSObject",
        "+ alloc;",
        "Class isa = 0x",
        "@property (readonly,copy) NSString * description;",
    ],
)
def test_symbol_without_super_str(hilda_client, sub_str: str):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param sub_str: Substring that should appear in the str.
    """
    assert str(hilda_client.objc_get_class("NSObject").new().objc_symbol).count(sub_str) == 1


@pytest.mark.parametrize(
    "sub_str",
    [
        "@interface",  # Class declaration.
        "NSMutableDictionary",  # Class type.
        "Class isa = 0x",  # Ivar.
        "@property (readonly,copy) NSString * description;",  # Property.
        "+ automaticallyNotifiesObserversForKey:(id);",  # Class method.
        "- removeAllObjects;",  # Instance method.
    ],
)
def test_symbol_str(hilda_client, sub_str: str):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param sub_str: Substring that should appear in the str.
    """
    assert hilda_client.ns({1: 2, 3: 4}).objc_symbol._to_str().count(sub_str) == 1


def test_symbol_str_not_recursive(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    assert "+ alloc;" not in hilda_client.ns({1: 2, 3: 4}).objc_symbol._to_str(False)


def test_symbol_str_recursive(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    assert "+ alloc;" in hilda_client.ns({1: 2, 3: 4}).objc_symbol._to_str(True)


@pytest.mark.skip("Not fully supported")
def test_set_implementation(hilda_client):
    pid = hilda_client.symbols.getpid()

    hilda_client.objc_get_class("NSJSONSerialization").methods.get("isValidJSONObject:").set_implementation(
        hilda_client.symbols.getpid
    )
    assert hilda_client.objc_get_class("NSJSONSerialization").isValidJSONObject_() == pid



from typing import Optional

from hilda.lldb_importer import lldb

_FILENAME = "/tmp/hilda-keylog.txt"


def _ssl_log_secret_bp(hilda, *args):
    label = hilda.registers.x1.peek_str()
    secret = hilda.registers.x2.peek(hilda.registers.x3).hex()
    random = (hilda.registers.x0[6] + 48).peek(32).hex()
    print(f"ssl_log_secret\n    label: {label}\n    secret: {secret}\n    random: {random}\n---\n")
    with open(_FILENAME, "a") as f:
        f.write(f"{label} {random} {secret}\n")
    hilda.cont()


def start_keylog(filename: Optional[str] = None) -> None:
    global _FILENAME

    if filename is not None:
        _FILENAME = filename
    hilda_client = lldb.hilda_client
    hilda_client.symbols._ZN4bssl14ssl_log_secretEPK6ssl_stPKcNS_4SpanIKhEE.bp(_ssl_log_secret_bp)



from construct import (
    Array,
    Bytes,
    Enum,
    Hex,
    Int8ul,
    Int32ul,
    Int64ul,
    PaddedString,
    Pointer,
    Seek,
    Struct,
    Switch,
    Tell,
    this,
)

from hilda.lldb_importer import lldb
from hilda.snippets.macho.apple_version import version_t
from hilda.symbol import SymbolFormatField

# See: https://opensource.apple.com/source/xnu/xnu-7195.81.3/EXTERNAL_HEADERS/mach-o/loader.h
# And: https://github.com/blacktop/go-macho/blob/master/types/commands.go
LC_REQ_DYLD = 0x80000000
LC_SEP = 0x8000000

LOAD_COMMAND_TYPE = Enum(
    Int32ul,
    LC_SEGMENT=0x1,
    LC_SYMTAB=0x2,
    LC_SYMSEG=0x3,
    LC_THREAD=0x4,
    LC_UNIXTHREAD=0x5,
    LC_LOADFVMLIB=0x6,
    LC_IDFVMLIB=0x7,
    LC_IDENT=0x8,
    LC_FVMFILE=0x9,
    LC_PREPAGE=0xA,
    LC_DYSYMTAB=0xB,
    LC_LOAD_DYLIB=0xC,
    LC_ID_DYLIB=0xD,
    LC_LOAD_DYLINKER=0xE,
    LC_ID_DYLINKER=0xF,
    LC_PREBOUND_DYLIB=0x10,
    LC_ROUTINES=0x11,
    LC_SUB_FRAMEWORK=0x12,
    LC_SUB_UMBRELLA=0x13,
    LC_SUB_CLIENT=0x14,
    LC_SUB_LIBRARY=0x15,
    LC_TWOLEVEL_HINTS=0x16,
    LC_PREBIND_CKSUM=0x17,
    LC_LOAD_WEAK_DYLIB=(0x18 | LC_REQ_DYLD),
    LC_SEGMENT_64=0x19,
    LC_ROUTINES_64=0x1A,
    LC_UUID=0x1B,
    LC_RPATH=(0x1C | LC_REQ_DYLD),
    LC_CODE_SIGNATURE=0x1D,
    LC_SEGMENT_SPLIT_INFO=0x1E,
    LC_REEXPORT_DYLIB=0x1F | LC_REQ_DYLD,
    LC_LAZY_LOAD_DYLIB=0x20,
    LC_ENCRYPTION_INFO=0x21,
    LC_DYLD_INFO=0x22,
    LC_DYLD_INFO_ONLY=(0x22 | LC_REQ_DYLD),
    LC_LOAD_UPWARD_DYLIB=(0x23 | LC_REQ_DYLD),
    LC_VERSION_MIN_MACOSX=0x24,
    LC_VERSION_MIN_IPHONEOS=0x25,
    LC_FUNCTION_STARTS=0x26,
    LC_DYLD_ENVIRONMENT=0x27,
    LC_MAIN=(0x28 | LC_REQ_DYLD),
    LC_DATA_IN_CODE=0x29,
    LC_SOURCE_VERSION=0x2A,
    LC_DYLIB_CODE_SIGN_DRS=0x2B,
    LC_ENCRYPTION_INFO_64=0x2C,
    LC_LINKER_OPTION=0x2D,
    LC_LINKER_OPTIMIZATION_HINT=0x2E,
    LC_VERSION_MIN_TVOS=0x2F,
    LC_VERSION_MIN_WATCHOS=0x30,
    LC_NOTE=0x31,
    LC_BUILD_VERSION=0x32,
    LC_DYLD_EXPORTS_TRIE=(0x33 | LC_REQ_DYLD),
    LC_DYLD_CHAINED_FIXUPS=(0x34 | LC_REQ_DYLD),
    LC_FILESET_ENTRY=(0x35 | LC_REQ_DYLD),
    LC_ATOM_INFO=0x36,
    LC_FUNCTION_VARIANTS=0x37,
    LC_FUNCTION_VARIANT_FIXUPS=0x38,
    LC_TARGET_TRIPLE=0x39,
    LC_SEP_CACHE_SLIDE=(0x1 | LC_SEP),
)


def __lc_str_from_load_command(this):
    load_command = this._._._
    return Struct(
        # this struct appears as a union in apple's opensource, but is represented in an awful manner, whereas
        # they would treat both the `ptr` and `offset` members as offsets (and never as pointers),
        # so we changed this a bit into the struct that it should have been
        "offset" / Hex(Int32ul),
        "_pad" / Int32ul,
        "name" / Pointer(load_command._start + this.offset, PaddedString(load_command.cmdsize - this.offset, "utf8")),
    )


__dylib_t = Struct(
    "lc_str" / __lc_str_from_load_command(this),
    "timestamp" / Int64ul,
    "current_version" / Int64ul,
    "compatibility_version" / Int64ul,
)

# Load commands:
# reference - https://opensource.apple.com/source/xnu/xnu-2050.18.24/EXTERNAL_HEADERS/mach-o/loader.h
__dylib_command_t = Struct(
    "dylib" / __dylib_t,
)

__segment_command_t = Struct(
    "segname" / PaddedString(16, "utf8"),
    "vmaddr" / SymbolFormatField(lldb.hilda_client),
    "vmsize" / Int64ul,
    "fileoff" / Int64ul,
    "filesize" / Int64ul,
    "maxprot" / Int32ul,
    "initprot" / Int32ul,
    "nsects" / Int32ul,
    "flags" / Int32ul,
)

__uuid_command_t = Struct("uuid" / Array(16, Int8ul))

__build_tool_version = Struct("tool" / Int32ul, "version" / Int32ul)

__build_version_command_t = Struct(
    "platform" / Int32ul,
    "minos" / version_t,
    "sdk" / version_t,
    "ntools" / Int32ul,
    # '_build_tools' / SymbolFormatField(this),
    # 'build_tools' / If(this._build_tools != 0, Array(this.ntools, Pointer(this._build_tools, __build_tool_version)))
)

load_command_t = Struct(
    "_start" / Tell,
    "cmd" / LOAD_COMMAND_TYPE,
    "cmdsize" / Int32ul,
    "_data_offset" / Tell,
    "data"
    / Switch(
        this.cmd,
        {
            LOAD_COMMAND_TYPE.LC_BUILD_VERSION: __build_version_command_t,
            LOAD_COMMAND_TYPE.LC_UUID: __uuid_command_t,
            LOAD_COMMAND_TYPE.LC_SEGMENT_64: __segment_command_t,
            LOAD_COMMAND_TYPE.LC_LOAD_DYLIB: __dylib_command_t,
            LOAD_COMMAND_TYPE.LC_LOAD_WEAK_DYLIB: __dylib_command_t,
            LOAD_COMMAND_TYPE.LC_ID_DYLIB: __dylib_command_t,
            LOAD_COMMAND_TYPE.LC_REEXPORT_DYLIB: __dylib_command_t,
            LOAD_COMMAND_TYPE.LC_LAZY_LOAD_DYLIB: __dylib_command_t,
        },
        Bytes(this.cmdsize - (this._data_offset - this._start)),
    ),
    Seek(this._start + this.cmdsize),
)


class LoadCommand:
    def __init__(self, load_command_data):
        self.__load_command_data = load_command_data
        self.__cmd = load_command_data.cmd
        self.__cmd_size = load_command_data.cmdsize
        self.__data_offset = load_command_data._data_offset

    @property
    def cmd(self):
        return self.__cmd

    @property
    def cmd_size(self):
        return self.__cmd_size


class DylibCommand(LoadCommand):
    def __init__(self, load_command_data):
        super().__init__(load_command_data)
        dylib_data = load_command_data.data.dylib

        self.__path = dylib_data.lc_str.name
        self.__timestamp = dylib_data.timestamp
        self.__current_version = dylib_data.current_version
        self.__compatibility_version = dylib_data.compatibility_version

    def __str__(self):
        return f"<DylibCommand: path = {self.path}>"

    def __repr__(self):
        return self.__str__()

    @property
    def path(self):
        return self.__path


class Segment64Command(LoadCommand):
    def __init__(self, load_command_data):
        super().__init__(load_command_data)

        self.__segname = load_command_data.data.segname
        self.__vmaddr = load_command_data.data.vmaddr
        self.__vmsize = load_command_data.data.vmsize
        self.__fileoff = load_command_data.data.fileoff
        self.__filesize = load_command_data.data.filesize
        self.__maxprot = load_command_data.data.maxprot
        self.__initprot = load_command_data.data.initprot
        self.__nsects = load_command_data.data.nsects
        self.__flags = load_command_data.data.flags

    @property
    def vmaddr(self):
        return self.__vmaddr

    @property
    def segname(self):
        return self.__segname

    def __str__(self):
        return f"<Segment64Command: segname={self.segname}, vmaddr={self.vmaddr}, vmsize={self.__vmsize}>"

    def __repr__(self):
        return self.__str__()


class UUIDCommand(LoadCommand):
    def __init__(self, load_command_data):
        super().__init__(load_command_data)

        self.__uuid = load_command_data.data.uuid

    def __str__(self):
        return "<UUIDCommand>"

    def __repr__(self):
        return self.__str__()


class BuildVersionCommand(LoadCommand):
    def __init__(self, load_command_data):
        super().__init__(load_command_data)

        self.__platform = load_command_data.data.platform
        self.__minos = load_command_data.data.platform
        self.__sdk = load_command_data.data.platform
        self.__ntools = load_command_data.data.platform
        self.__build_tools = load_command_data.data.build_tools

    def __str__(self):
        return f"<BuildVersionCommand platform={self.platform}>"

    def __repr__(self):
        return self.__str__()

    @property
    def platform(self):
        return self.__platform

    @property
    def minos(self):
        return self.__minos

    @property
    def sdk(self):
        return self.__sdk


class UnimplementedCommand(LoadCommand):
    def __init__(self, load_command_data):
        super().__init__(load_command_data)

        self.__bytes = load_command_data.data

    def __str__(self):
        return f"<cmd={self.cmd} is not implemented((Feel free to add!!). bytes={self.__bytes}>"

    def __repr__(self):
        return self.__str__()


class LoadCommands:
    def __init__(self, load_commands_data):
        self.__load_commands = []
        for load_command_data in load_commands_data:
            if load_command_data.cmd == "LC_SEGMENT_64":
                load_command = Segment64Command(load_command_data)
            elif (
                load_command_data.cmd == "LC_LOAD_DYLIB"
                or load_command_data.cmd == "LC_ID_DYLIB"
                or load_command_data.cmd == "LC_LOAD_WEAK_DYLIB"
                or load_command_data.cmd == "LC_REEXPORT_DYLIB"
                or load_command_data.cmd == "LC_LAZY_LOAD_DYLIB"
            ):
                load_command = DylibCommand(load_command_data)
            elif load_command_data.cmd == "LC_UUID":
                load_command = UUIDCommand(load_command_data)
            else:
                load_command = UnimplementedCommand(load_command_data)

            self.__load_commands.append(load_command)

    @property
    def all(self):
        return self.__load_commands

    @property
    def segment_commands(self) -> list[Segment64Command]:
        return [
            segment_command for segment_command in self.__load_commands if isinstance(segment_command, Segment64Command)
        ]

    @property
    def dylib_commands(self) -> list[DylibCommand]:
        return [dylib_command for dylib_command in self.__load_commands if isinstance(dylib_command, DylibCommand)]

    def find(self, predicate=None) -> list[LoadCommand]:
        if predicate is None:
            return self.__load_commands

        matching_load_commands = []
        for load_command in self.__load_commands:
            if predicate(load_command):
                matching_load_commands.append(load_command)

        return matching_load_commands



import logging

SYSLOG_LISTENER_CODE = """
socat UNIX-RECVFROM:/var/run/syslog, reuseaddr, fork, mode-0777 STDOUT
"""


def open_syslog_socket():
    logging.info(f"please execute the following code on the device for log listening:\n{SYSLOG_LISTENER_CODE}")
    input("> Hit return to resume")



__all__ = [
    "AccessingMemoryError",
    "AccessingRegisterError",
    "AddingLldbSymbolError",
    "ConvertingFromNSObjectError",
    "ConvertingToNsObjectError",
    "CreatingObjectiveCSymbolError",
    "DisableJetsamMemoryChecksError",
    "EvaluatingExpressionError",
    "GettingObjectiveCClassError",
    "HildaException",
    "InvalidThreadIndexError",
    "LLDBError",
    "SymbolAbsentError",
]


class HildaException(Exception):
    """A domain exception for hilda errors."""

    pass


class LLDBError(HildaException):
    """Wrapper for RAW LLDB errors"""

    pass


class SymbolAbsentError(HildaException):
    """Raise when trying to get a symbol that doesn't exist."""

    pass


class EvaluatingExpressionError(HildaException):
    """Raise when failing to evaluate an expression."""

    pass


class CreatingObjectiveCSymbolError(HildaException):
    """Raise when failing to create an ObjectiveC Symbol."""

    pass


class ConvertingToNsObjectError(HildaException):
    """Raise when failing to convert python object to NS object."""

    pass


class ConvertingFromNSObjectError(HildaException):
    """Raise when failing to convert NS object to python object."""

    pass


class DisableJetsamMemoryChecksError(HildaException):
    """Raise when failing to disable jetsam memory checks."""

    pass


class GettingObjectiveCClassError(HildaException):
    """Raise when failing to get an ObjectiveC class."""

    pass


class AccessingRegisterError(HildaException):
    """Raise when failing to access a register."""

    pass


class AccessingMemoryError(HildaException):
    """Raise when failing to access memory."""

    pass


class AddingLldbSymbolError(HildaException):
    """Raise when failing to convert a LLDB symbol to Hilda's symbol."""

    pass


class InvalidThreadIndexError(HildaException):
    """Raise when thread idx invalid"""

    pass



from construct import Array, Int8ul, Int32ul, Int64ul, Struct

uuid_t = Struct(
    "time_low" / Int64ul,
    "time_mid" / Int32ul,
    "time_hi_and_version" / Int32ul,
    "clock_seq_hi_and_reserved" / Int8ul,
    "clock_seq_low" / Int8ul,
    "node" / Array(6, Int8ul),
)
